
# Example code extracted automatically from documentation.
# Do not edit these files manually.
from ._017C_segmentation_males import example as _017C_segmentation_males
from ._017A_market_seg import example as _017A_market_seg
from ._016_cost_by_income import example as _016_cost_by_income
from ._003_zeroed_shared import example as _003_zeroed_shared
from ._017D_segmentation_female import example as _017D_segmentation_female
from ._013_cbd import example as _013_cbd
from ._017_mnl import example as _017_mnl
from ._014_work_zone_density import example as _014_work_zone_density
from ._022_nl import example as _022_nl
from ._007_ovtt_by_distance import example as _007_ovtt_by_distance
from ._017B_alt_segmentation import example as _017B_alt_segmentation
from ._012_autos_per_adult import example as _012_autos_per_adult
from ._009_ttr_4 import example as _009_ttr_4
from ._030_mnl_constrain import example as _030_mnl_constrain
from ._006_motorized_times import example as _006_motorized_times
from ._015_cbd_and_work_density import example as _015_cbd_and_work_density
from ._011_veh_per_worker import example as _011_veh_per_worker
from ._005_motorized_travel_time import example as _005_motorized_travel_time
from ._010_auto_per_hh import example as _010_auto_per_hh
from ._008_ttr import example as _008_ttr
from ._002_motorized import example as _002_motorized
from ._001_mnl import example as _001_mnl
from ._102_swissmetro_weighted import example as _102_swissmetro_weighted
from ._151_swissmetro_panel_latent_class import example as _151_swissmetro_panel_latent_class
from ._109_swissmetro_nl import example as _109_swissmetro_nl
from ._101_swissmetro_mnl import example as _101_swissmetro_mnl
from ._200_exampville import example as _200_exampville
from ._251_bySpec import example as _251_bySpec
from ._202_exville_mc_logsums import example as _202_exville_mc_logsums
from ._203_exville_dest import example as _203_exville_dest
from ._201_exville_mode_choice import example as _201_exville_mode_choice
from ._204_exville_dest_only import example as _204_exville_dest_only
from ._xlogit_test import example as _xlogit_test

ex = {}
ex['017C'] = _017C_segmentation_males
ex['017A'] = _017A_market_seg
ex[16] = _016_cost_by_income
ex[3] = _003_zeroed_shared
ex['017D'] = _017D_segmentation_female
ex[13] = _013_cbd
ex[17] = _017_mnl
ex[14] = _014_work_zone_density
ex[22] = _022_nl
ex[7] = _007_ovtt_by_distance
ex['017B'] = _017B_alt_segmentation
ex[12] = _012_autos_per_adult
ex[9] = _009_ttr_4
ex[30] = _030_mnl_constrain
ex[6] = _006_motorized_times
ex[15] = _015_cbd_and_work_density
ex[11] = _011_veh_per_worker
ex[5] = _005_motorized_travel_time
ex[10] = _010_auto_per_hh
ex[8] = _008_ttr
ex[2] = _002_motorized
ex[1] = _001_mnl
ex[102] = _102_swissmetro_weighted
ex[151] = _151_swissmetro_panel_latent_class
ex[109] = _109_swissmetro_nl
ex[101] = _101_swissmetro_mnl
ex[200] = _200_exampville
ex[251] = _251_bySpec
ex[202] = _202_exville_mc_logsums
ex[203] = _203_exville_dest
ex[201] = _201_exville_mode_choice
ex[204] = _204_exville_dest_only
ex['xlogit'] = _xlogit_test
